/* Copyright (C) 2009 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <stdio.h>

#include <alloca.h>

#include "huffman.h"

/* Gives the number of code words of a given bit length, the first
   code word, and the index of that code word. Code words are selected
   and ordered according to RFC 1951 (lexicographic order, shorter
   codes precede longer, codes of the same length are consecutive). */
struct code_info {
  unsigned short count;
  unsigned short index;
  unsigned short code;
};

/* Construct a huffman tree representing the codes starting with the
   given prefix. Requires that the prefix matches at least one code
   word. Returns the number of used nodes. */
static unsigned
huffman_build_tree(struct huffman_node *tree,
		   unsigned bits,
		   /* START[k] gives the first code word of length k. */
		   const struct code_info *info,
		   /* VALUE[k] gives the value for the k'th code word. */
		   const unsigned short *value,
		   unsigned prefix_length, unsigned prefix)
{
  /* Find the largest code word matching the prefix. */
  for (; bits >= prefix_length; bits--)
    {
      /* The largest code word of length bits that matches the prefix. */
      unsigned prefix_max;

      assert (bits > 0);
      if (!info[bits].count)
	continue;

      prefix_max = ((prefix+1) << (bits - prefix_length)) - 1;

      if (info[bits].code <= prefix_max)
	{
	  unsigned index;
	  unsigned code;

	  /* Index relative to the first code of this length. */
	  index = prefix_max - info[bits].code;
	  if (index >= info[bits].count)
	    index = info[bits].count - 1;

	  code = info[bits].code + index;

	  assert ((code >> (bits - prefix_length)) == prefix);
	  
	  /* We have the largest code word matching the prefix. If the next
	     bit of this code word is zero, that means that there's no choice,
	     the zero is mandatory. */
	  tree->zeros = 0;
	  while (prefix_length < bits
		 && !((code >> (bits - prefix_length - 1)) & 1))
	    {
	      tree->zeros++;
	      prefix_length++;
	      prefix <<= 1;
	    }

	  if (bits == prefix_length)
	    {
	      /* A leaf node */
	      tree->leaf = 1;
	      tree->value = value[info[bits].index + index];
	      return 1;
	    }
	  else
	    {
	      unsigned left
		= huffman_build_tree(tree + 1,
				     bits, info, value,
				     prefix_length + 1,
				     prefix << 1);
	      unsigned right
		= huffman_build_tree(tree + 1 + left,
				     bits, info, value,
				     prefix_length + 1,
				     (prefix << 1) + 1);

	      tree->leaf = 0;
	      tree->value = 1 + left;
	      return 1 + left + right;
	    }
	}
    }
  abort ();
}

#define MAX_LENGTH 15

int
huffman_tree(struct huffman_node *tree,
	     unsigned n, const unsigned char *length)
{
  struct code_info info[MAX_LENGTH + 1];
  unsigned short index[MAX_LENGTH + 1];

  unsigned k;
  unsigned bits;
  /* Gives the value for the j'th code word, with code words ordered
     lexicographically. */
  unsigned short *value;
  unsigned size;

  memset(info, 0, sizeof(info));

  for (k = 0; k < n; k++)
    {
      if (length[k] > MAX_LENGTH)
	return 0;
      info[length[k]].count++;
    }

  /* RFC 1951 seem to have an off-by-one error. */
  info[1].code = 0;
  info[1].index = 0;

  for (bits = 2; bits <= MAX_LENGTH; bits++)
    {
      info[bits].code = (info[bits-1].code + info[bits-1].count) << 1;
      info[bits].index = info[bits-1].index + info[bits-1].count;

      /* Make sure code fits in given size. */
      if (info[bits].code + info[bits].count > (1U << bits))
	return 0;
    }
  
  value = alloca((n - info[0].count)  * sizeof(*value));
  memset(index, 0, sizeof(index));

  for (k = 0; k < n; k++)
    if (length[k] > 0)
      value[info[length[k]].index + index[length[k]]++] = k;
    
  size = huffman_build_tree(tree, MAX_LENGTH, info, value, 0, 0);
  assert(size <= HUFFMAN_SIZE(n));
  assert(size == HUFFMAN_SIZE(n - info[0].count));

  /* Sanity check */
  for (k = 0; k < size; k++)
    assert(tree[k].leaf || k + tree[k].value < size);

  return 1;
}

void
huffman_init (struct huffman_ctx *ctx,
	      const struct huffman_node *root)
{
  assert(!root->leaf ||root->zeros);

  ctx->root = root;
  ctx->p = root;
  ctx->zeros = root->zeros;
}

int
huffman_decode_bit (struct huffman_ctx *ctx, unsigned bit, unsigned *value)
{
  if (ctx->zeros > 0)
    {
      /* Next bit must be zero. */
      if (bit)
	return -1;
      
      ctx->zeros--;
    }
  else
    {
      assert(!ctx->p->leaf);

      ctx->p += bit ? ctx->p->value : 1;
      ctx->zeros = ctx->p->zeros;
    }

  if (ctx->zeros > 0)
    return 0;

  else if (ctx->p->leaf)
    {
      *value = ctx->p->value;

      ctx->p = ctx->root;
      ctx->zeros = ctx->root->zeros;

      return 1;
    }
  else
    return 0;
}
